# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeAlias
from typing import TypeVar

from .iclient import IClient
from .iclientauthorizationstate import IClientAuthorizationState
from .iobtainedcredential import IObtainedCredential


C = TypeVar('C', bound=IClient)
S = TypeVar('S', bound=IClientAuthorizationState)
T = TypeVar('T', bound=IObtainedCredential)


class IClientStorage(Generic[C, S, T]):
    __module__: str = 'oauthx.client.protocols'

    async def delete(self, obj: C | S| T) -> None: ...
    async def get_client(self, client_id: str) -> C: ...
    async def get_credential(self, client_id: str, resource: str | None = None) -> T | None: ... 
    async def get_state(self, state: str) -> S | None: ...
    async def persist(self, obj: C | S | T) -> None: ...


ClientStorageType: TypeAlias = IClientStorage[
    IClient,
    IClientAuthorizationState,
    IObtainedCredential
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from oauthx.lib import RFC9068AccessToken
from oauthx.lib import ServerMetadata
from oauthx.lib.types import AccessTokenHash


class Request(fastapi.Request):
    access_token: RFC9068AccessToken
    at_hash: AccessTokenHash
    issuer: ServerMetadata


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from oauthx.lib.exceptions import InteractionException


class InvalidResponseType(InteractionException):
    """Raised when the `response_type` URI is invalid or not
    allowed by the client.
    """
    __module__: str = 'oauthx.server.types'
    error: str = 'unsupported_response_type'
    status_code = 403
    template_name = 'oauthx/errors/authorize.redirect.html.j2'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
import pytest
from aiopki.ext import jose

from oauthx.client import Client as ConsumingClient
from oauthx.lib import RFC9068AccessToken
from oauthx.lib.exceptions import Error
from .types import AuthorizeFactory


@pytest.fixture(scope='function')
def grant_types() -> list[str]:
    return ['authorization_code', 'refresh_token']


@pytest.fixture(scope='function')
def scope() -> list[str]:
    return ['offline_access', 'email', 'profile']


@pytest.mark.asyncio
async def test_excessive_scope_raises_invalid_scope(
    agent: httpx.AsyncClient,
    client: ConsumingClient,
    authorize_request: AuthorizeFactory
):
    _, response, state = await authorize_request(
        scope={'offline_access', 'email'}
    )
    assert response
    assert state
    grant = await client.authorization_code(response, state, http=agent)
    assert grant.refresh_token is not None
    try:
        await client.refresh(grant.refresh_token, scope={'email', 'profile'}, http=agent)
        assert False
    except Error as exc:
        assert exc.error == 'invalid_scope'


@pytest.mark.asyncio
async def test_omitted_scope_grants_original_scope(
    agent: httpx.AsyncClient,
    client: ConsumingClient,
    authorize_request: AuthorizeFactory
):
    _, response, state = await authorize_request(
        scope={'offline_access', 'email'}
    )
    assert response
    assert state
    grant = await client.authorization_code(response, state, http=agent)
    assert grant.refresh_token is not None
    refreshed = await client.refresh(grant.refresh_token, http=agent)
    jws = jose.parse(refreshed.access_token)
    at = jws.payload(RFC9068AccessToken.model_validate)
    assert set(at.scope) == {'offline_access', 'email'}


@pytest.mark.asyncio
async def test_original_scope_subset(
    agent: httpx.AsyncClient,
    client: ConsumingClient,
    authorize_request: AuthorizeFactory
):
    _, response, state = await authorize_request(
        scope={'offline_access', 'email'}
    )
    assert response
    assert state
    grant = await client.authorization_code(response, state, http=agent)
    assert grant.refresh_token is not None
    refreshed = await client.refresh(grant.refresh_token, scope={'email'}, http=agent)
    jws = jose.parse(refreshed.access_token)
    at = jws.payload(RFC9068AccessToken.model_validate)
    assert set(at.scope) == {'email'}


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Iterable
from typing import Literal

import httpx

from oauthx.client.models import Client
from oauthx.lib.models import ObtainedGrant
from oauthx.client.auth import LocalAuthorization


class GoogleCommandLineAuth(LocalAuthorization):
    use: Literal['access_token', 'id_token']

    def __init__(
        self,
        scope: Iterable[str] | None = None,
        path: str | None = None,
        use: Literal['access_token', 'id_token'] = 'access_token'
    ):
        client = Client.model_validate({
            'provider': 'https://accounts.google.com',
            'client_id': '764086051850-6qr4p6gpi6hn506pt8ejuq83di341hur.apps.googleusercontent.com',
            'credential': 'd-FL95Q19q7MQmFpd7hHD0Ty',
            'token_endpoint_auth_method': 'client_secret_basic'
        })
        scope = set(scope or [])
        scope.update(['email'])
        super().__init__(client, scope, path)
        self.use = use

    def get_token(self, grant: ObtainedGrant) -> str:
        token = grant.access_token
        if self.use == 'id_token':
            assert grant.id_token is not None
            token = grant.id_token.encode(bytes.decode, compact=True)
        return token


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from oauthx.lib.protocols import BaseSettings


class ClientSettings(BaseSettings):
    splash_image: str = 'https://static.webiam.id/shared/splash.jpg'

    def get_splash_image_url(self) -> str:
        return self.splash_image


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

import fastapi


class IPlugin(Protocol):
    __module__: str = 'oauthx.server.protocols'

    @classmethod
    def __register__(cls, app: fastapi.FastAPI | fastapi.APIRouter) -> None:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import MutableMapping

from oauthx.lib.protocols import IRegisteredClient


class UnregisteredClient(IRegisteredClient):

    @property
    def id(self) -> str:
        raise NotImplementedError

    def contribute_to_event(self, data: MutableMapping[str, Any]) -> None:
        pass

    def get_display_name(self) -> str:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.protocols import BaseSettings


__all__: list[str] = [
    'CurrentSettings',
    'CurrentSettingsDependency',
]


def get(request: fastapi.Request) -> BaseSettings:
    return request.config # type: ignore



CurrentSettingsDependency: BaseSettings = fastapi.Depends(get)
CurrentSettings: TypeAlias = Annotated[BaseSettings, CurrentSettingsDependency]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
import pytest
from aiopki.ext import jose

from oauthx.client import Client as ConsumingClient
from oauthx.lib import RFC9068AccessToken
from oauthx.lib.exceptions import Error
from .types import AuthorizeFactory


@pytest.fixture(scope='function')
def grant_types() -> list[str]:
    return ['authorization_code', 'refresh_token']


@pytest.fixture(scope='function')
def scope() -> list[str]:
    return ['offline_access', 'email', 'profile']


@pytest.mark.asyncio
@pytest.mark.parametrize("resource", [
    ['https://example.com', 'https://example.net'],
    'https://example.com',
])
async def test_access_token_must_be_restricted_to_the_audience(
    agent: httpx.AsyncClient,
    client: ConsumingClient,
    resource: str | list[str],
    authorize_request: AuthorizeFactory
):
    _, response, state = await authorize_request(resource=resource)
    assert not response.is_error()
    grant = await client.authorization_code(response, state, http=agent)
    jws = jose.parse(grant.access_token)
    at = jws.payload(RFC9068AccessToken.model_validate)
    if not isinstance(resource, list):
        resource = [resource]
    assert set(at.aud or []) == set(resource)


@pytest.mark.parametrize("resource", [
    'https://example.com',
    ['https://example.com', 'https://example.net']
])
@pytest.mark.asyncio
async def test_can_not_obtain_token_with_non_consented_resources(
    agent: httpx.AsyncClient,
    client: ConsumingClient,
    resource: str | list[str],
    authorize_request: AuthorizeFactory
):
    _, params, state = await authorize_request(resource=resource)
    assert not params.is_error()
    try:
        await client.authorization_code(
            params=params,
            state=state,
            resources='https://www.foo.com',
            http=agent
        )
        pytest.fail()
    except Error as exc:
        assert exc.error == 'invalid_target'


@pytest.mark.parametrize("resource", [
    'https://example.com',
    ['https://example.com', 'https://example.net']
])
@pytest.mark.asyncio
async def test_can_not_refresh_token_with_non_consented_resources(
    agent: httpx.AsyncClient,
    client: ConsumingClient,
    resource: str | list[str],
    authorize_request: AuthorizeFactory
):
    _, params, state = await authorize_request(resource=resource, scope={'offline_access'})
    assert not params.is_error()
    obtained = await client.authorization_code(
        params=params,
        state=state,
        http=agent
    )
    assert obtained.refresh_token is not None
    try:
        await client.refresh(
            obtained.refresh_token,
            resource='https://example.org',
            http=agent
        )
        pytest.fail()
    except Error as exc:
        assert exc.error == 'invalid_target'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Annotated
from typing import TypeAlias

import fastapi
from aiopki.utils import b64decode


__all__: list[str] = ['TelegramToken']


def get() -> bytes:
    return b64decode(os.environ['TELEGRAM_LOGIN_TOKEN'])


TelegramToken: TypeAlias = Annotated[bytes, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from aiopki.ext import jose
from aiopki.ext.jose import JWT
from aiopki.types import ISigner

from oauthx.server.types import AuthorizationKey
from .clientkey import ClientKey
from .resourceownerkey import ResourceOwnerKey


class RefreshToken(JWT):
    aut: int = pydantic.Field(
        default=...
    )

    client_id: ClientKey = pydantic.Field(
        default=...
    )

    @property
    def authorization(self) -> AuthorizationKey:
        return AuthorizationKey(self.aut)

    @property
    def owner(self) -> ResourceOwnerKey:
        assert self.sub is not None
        return ResourceOwnerKey(str(self.client_id), self.sub)

    async def sign(self, signer: ISigner) -> str:
        jws = jose.jws(self.model_dump(exclude_none=True))
        await jws.sign(
            algorithm=signer.default_algorithm(),
            signer=signer,
            protected={'typ': 'jwt+refresh-token'}
        )
        return jws.encode(compact=True, encoder=bytes.decode)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from oauthx.client.models import Client
from oauthx.client.models import ClientAuthorizationState
from oauthx.client.models import ObtainedCredential
from oauthx.client.protocols import IClientStorage
from oauthx.client import LocalClientRepository

from .basedatastorestorage import BaseDatastoreStorage


class ClientRepository(
    IClientStorage[Client, ClientAuthorizationState, ObtainedCredential],
    BaseDatastoreStorage
):
    local: LocalClientRepository = LocalClientRepository()

    async def get_client(self, client_id: str):
        return await self.local.one(client_id)

    async def get_credential(
        self,
        client_id: str,
        resource: str | None = None
    ) -> ObtainedCredential | None:
        return await self.get_model_by_key(ObtainedCredential, client_id)

    async def get_state(self, state: str) -> ClientAuthorizationState | None:
        return await self.get_model_by_key(ClientAuthorizationState, state)

    async def persist(self, obj: Client | ClientAuthorizationState | ObtainedCredential) -> None:
        await self.put(obj, obj.pk)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi

from oauthx.server.models import SubjectKey
from oauthx.server.params import AuthorizationServer
from oauthx.server.params import CurrentSubject
from oauthx.server.params import ObjectFactory
from oauthx.server.params import PluginRunner
from oauthx.server.params import RequestSession
from oauthx.server.request import Request
from ..baserequesthandler import BaseRequestHandler
from .params import AuthorizationRequest
from .params import Client
from .params import ResponseMode
from .params import ResourceOwner
from .params import UserInfo



router: fastapi.APIRouter = fastapi.APIRouter()


@router.options('/authorize', name='oauth2.authorize')
async def options(
    client: Client
):
    raise NotImplementedError


@router.get('/authorize', name='oauth2.authorize')
async def get(
    server: AuthorizationServer,
    factory:  ObjectFactory,
    plugins: PluginRunner,
    client: Client,
    mode: ResponseMode,
    params: AuthorizationRequest,
    session: RequestSession,
    userinfo: UserInfo,
) -> fastapi.Response:
    response = await plugins.validate_scope(
        client=client,
        request=params,
        scope=params.scope
    )
    if response is not None:
        return response

    async with params.consume():
        query = await server.authorize(
            client=client,
            userinfo=userinfo,
            params=params,
            mode=mode,
            contributors=[params, session]
        )
        return await mode.redirect(**query)
    authorization = await factory.authorization(
        request=params,
        client_id=client.id,
        lifecycle='GRANTED',
        scope=params.scope,
        sub=SubjectKey(userinfo.sub),
        token_types=mode.get_token_types(),
        contributors=[params, session]
    )
    await authorization.persist()
    await params.delete()
    return mode.redirect()


class AuthorizationRequestHandler(BaseRequestHandler):
    """Provides an interface for the resource owner to authorize a certain
    scope for a client, and redirect back to the clients' redirection
    endpoint.
    """
    __module__: str = 'oauthx.server.handlers'
    client: Client
    name: str = 'oauth2.authorize'
    path: str = '/authorize'
    responses: dict[int | str, Any] = {
        400: {
            'description': (
                "Unrecoverable error that is not allowed to redirect"
            )
        }
    }
    response_class: type[fastapi.Response] = fastapi.responses.RedirectResponse
    response_description: str = "Redirect to the clients' redirection endpoint."
    status_code: int = 302
    subject: CurrentSubject | None
    summary: str = "Authorization Endpoint"

    def setup(
        self,
        *,
        client: Client,
        response_mode: ResponseMode,
        subject: CurrentSubject,
        params: AuthorizationRequest,
        plugins: PluginRunner,
        owner: ResourceOwner,
        session: RequestSession,
        userinfo: UserInfo,
        **_: Any
    ):
        self.client = client
        self.owner = owner
        self.params = params
        self.plugins = plugins
        self.response_mode = response_mode
        self.session = session
        self.subject = subject
        self.userinfo = userinfo

    async def handle(self, request: Request) -> fastapi.Response:
        assert self.subject
        assert self.owner
        response = await self.plugins.validate_scope(
            client=self.client,
            request=self.params,
            scope=self.params.scope
        )
        if response is not None:
            return response

        assert self.params.id
        await self.storage.delete(self.params)
        authorization = await self.factory.authorization(
            request=self.params,
            client_id=self.client.id,
            lifecycle='GRANTED',
            scope=self.params.scope,
            sub=self.subject.get_primary_key(), # type: ignore
            token_types=self.response_mode.get_token_types(),
            contributors=[self.params, self.session]
        )
        await authorization.persist()

        async with self.owner.atomic():
            self.owner.consent(authorization.scope)

        return await self.response_mode.redirect(
            code=await self.server.issue_authorization_code(
                client=self.client,
                owner=self.owner,
                authorization_id=authorization.id,
                redirect_uri=self.params.redirect_uri
            )
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.server.models import Authorization as AuthorizationModel
from oauthx.server.models import SubjectKey
from oauthx.server.params import ObjectFactory
from oauthx.server.params import RequestSession
from oauthx.server.params import Storage
from .authorizationrequest import AuthorizationRequest
from .client import Client
from .responsemode import ResponseMode
from .userinfo import UserInfo


__all__: list[str] = [
    'Authorization'
]


async def get(
    factory: ObjectFactory,
    storage: Storage,
    session: RequestSession,
    client: Client,
    params: AuthorizationRequest,
    response_mode: ResponseMode,
    userinfo: UserInfo
) -> AuthorizationModel:
    authorization = await factory.authorization(
        request=params,
        client_id=client.id,
        lifecycle='GRANTED',
        scope=params.scope,
        sub=SubjectKey(userinfo.sub),
        token_types=response_mode.get_token_types(),
        contributors=[params, session]
    )
    await authorization.persist()
    await storage.delete(params)
    return authorization

Authorization: TypeAlias = Annotated[AuthorizationModel, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
import pydantic
from aiopki.ext.jose import CompactSerialized

from oauthx.lib.exceptions import InvalidGrant
from oauthx.lib.exceptions import InvalidRequest
from oauthx.server.params import Issuer
from oauthx.server.params import TokenSigner
from oauthx.server.models import RefreshToken as RefreshTokenModel
from .client import Client
from .granttype import GrantType
from .query import REFRESH_TOKEN


__all__: list[str] = [
    'RefreshToken'
]


MAX_AGE: int = 180 * 86400


async def get(
    issuer: Issuer,
    signer: TokenSigner,
    request: fastapi.Request,
    client: Client,
    grant_type: GrantType,
    jws: CompactSerialized | None = REFRESH_TOKEN
) -> RefreshTokenModel | None:
    if grant_type != 'refresh_token' or client is None:
        return None
    if jws is None:
        raise InvalidRequest("The refresh_token parameter is required.")
    try:
        rt = jws.payload(RefreshTokenModel.model_validate)
    except pydantic.ValidationError:
        raise InvalidGrant("The refresh token is malformed.")
    if not rt.validate_exp() or not rt.validate_nbf() or not rt.validate_iat(MAX_AGE):
        raise InvalidGrant("The refresh token is inactive or expired.")
    if not client.id == rt.client_id:
        raise InvalidGrant("The refresh token was issued to a different client.")
    if jws.typ != 'jwt+refresh-token':
        raise InvalidGrant("Invalid refresh token.")
    if not rt.validate_iss(issuer):
        raise InvalidGrant("The refresh token was issued by an untrusted issuer.")
    if not await jws.verify(signer):
        raise InvalidGrant("The signature of the refresh token did not validate.")
    if not rt.validate_aud(f'{request.url.scheme}://{request.url.netloc}{request.url.path}'):
        raise InvalidGrant(f"Invalid audience: {rt.aud}")
    return rt


RefreshToken: TypeAlias = Annotated[
    RefreshTokenModel | None,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Annotated
from typing import Any
from typing import Iterable

import fastapi
import fastapi.params
import pydantic
from aiopki.ext import jose
from aiopki.ext.jose import JWT
from aiopki.ext.jose import OIDCToken
from fastapi.exceptions import HTTPException
from fastapi.security import HTTPBearer
from fastapi.security import HTTPAuthorizationCredentials

from oauthx.client import Provider
from oauthx.lib.exceptions import TrustIssues
from oauthx.lib.params import HTTPClient


__all__: list[str] = ['IDTokenBearer']


PROVIDERS: dict[str, Provider] = {}
logger: logging.Logger = logging.getLogger('uvicorn')


def IDTokenBearer(
    issuer: str,
    required: bool = True,
    max_age: int | None = None,
    update_state: bool | set[str] | None = None,
    description: str | None = None,
    scheme_name: str = "OpenID Connect ID Token",
    audiences: set[str] | None = None,
    email: str | Iterable[str] | None = None,
    override: bool = False
):
    default_audiences = audiences or set()
    if isinstance(email, str):
        email = {email}
    allowed_email = set(email or [])
    issuers: set[str] = {issuer}
    security = HTTPBearer(auto_error=False, description=description, scheme_name=scheme_name)

    async def f(
        bearer: Annotated[HTTPAuthorizationCredentials | None, fastapi.Depends(security)],
        request: fastapi.Request, http: HTTPClient):
        nonlocal allowed_email
        nonlocal default_audiences
        nonlocal override
        nonlocal update_state
        audiences: set[str] = {
            f'{request.url.scheme}://{request.url.netloc}',
            f'{request.url.scheme}://{request.url.netloc}{request.url.path}',
            *default_audiences
        }
        if override and default_audiences:
            audiences = default_audiences
        if bearer is None:
            if required:
                raise HTTPException(
                    status_code=401,
                    detail=(
                        "Authentication is required for this endpoint. Provide "
                        "an OpenID Connect ID token using the Authorization "
                        "header and bearer scheme, issued by any of the following "
                        f"providers: {str.join(', ', sorted(issuers))}"
                    )
                )
            return None
        try:
            jws = jose.parse(bearer.credentials)
            id_token = jws.payload(JWT.model_validate)
        except Exception:
            raise HTTPException(
                status_code=401,
                detail="The Authorization header contained an invalid token."
            )
        if not any(map(id_token.validate_iss, issuers)):
            raise HTTPException(
                status_code=401,
                detail=(
                    'The JSON Web Token (JWT) provided in the Authorization '
                    'header was issued by an untrusted issuer. Trusted issuers are '
                    f'{str.join(", ", sorted(issuers))}.'
                )
            )

        if not id_token.validate_exp():
            raise HTTPException(
                status_code=401,
                detail=(
                    'The JSON Web Token (JWT) provided in the Authorization '
                    'header is expired.'
                )
            )

        if not id_token.validate_nbf(required=False):
            raise HTTPException(
                status_code=401,
                detail=(
                    'The JSON Web Token (JWT) provided in the Authorization '
                    'header is expired.'
                )
            )

        if max_age is not None and not id_token.validate_iat(max_age):
            raise HTTPException(
                status_code=401,
                detail=(
                    'The JSON Web Token (JWT) provided in the Authorization '
                    'header is too old or did not contain the "iat" claim.'
                )
            )

        if not id_token.validate_aud(audiences):
            raise HTTPException(
                status_code=401,
                detail=(
                    'The JSON Web Token (JWT) provided in the Authorization '
                    'header is not intended for this resource server.'
                )
            )

        assert id_token.iss is not None
        if id_token.iss not in PROVIDERS:
            logger.info("Retrieving issuer metadata (issuer: %s)", id_token.iss)
            provider = PROVIDERS[id_token.iss] = Provider.model_validate({
                'iss': id_token.iss
            })
            await provider.discover(http)

        provider = PROVIDERS[id_token.iss]
        if not provider.can_verify_id_token():
            raise HTTPException(
                status_code=503,
                detail=(
                    'The JSON Web Token (JWT) provided in the Authorization '
                    'header was properly validated but the issuer does not '
                    'provide public keys to verify the signature.'
                )
            )

        is_valid = False
        try:
            await provider.verify_id_token(jws)
            is_valid = True
        except TrustIssues:
            is_valid = False
        if not is_valid:
            raise HTTPException(
                status_code=403,
                detail=(
                    'The JSON Web Signature (JWS) provided in the Authorization '
                    'header did not verify against the known public keys for its '
                    'issuer.'
                )
            )
        try:
            id_token = OIDCToken.model_validate(id_token.model_dump())
            claims: dict[str, Any] = id_token.model_dump()

            if allowed_email and claims.get('email') not in allowed_email:
                raise HTTPException(
                    status_code=401,
                    detail=(
                        "The email address is not allowed to authenticate."
                    )
                )

            if update_state == True:
                update_state = set(claims.keys())
            missing: set[str] = set()
            for name in (update_state or set()):
                if name not in claims:
                    missing.add(name)
                    continue
                setattr(request.state, name, claims[name])
            if missing:
                raise HTTPException(
                    status_code=401,
                    detail=(
                        "The ID Token did not contain the required claims: "
                        f"{', '.join(sorted(missing))}."
                    )
                )
            setattr(request.state, 'id_token', id_token)
        except pydantic.ValidationError:
            raise HTTPException(
                status_code=403,
                detail=(
                    'The JSON Web Token (JWT) provided in the Authorization '
                    'header could be parsed and had a valid signature, but '
                    'is not a valid Open ID Connect ID Token.'
                )
            )
        return getattr(request.state, 'id_token')
            

    return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol

import httpx

from .iclient import IClient


class IObtainedCredential(Protocol):
    __module__: str = 'oauthx.client.protocols'

    @property
    def pk(self) -> str:
        ...

    def add_to_request(self, request: httpx.Request) -> None:
        ...

    async def destroy(self) -> None:
        ...

    async def refresh(self, client: IClient):
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol

import fastapi
from aiopki.types import ISigner

from oauthx.lib.types import GrantType
from oauthx.lib.types import ResponseType


class IResponseMode(Protocol):
    __module__: str = 'oauthx.server.types'
    response_type: ResponseType

    def get_token_types(self) -> set[str]: ...
    def grants(self) -> GrantType | None: ...
    def with_signer(self, signer: ISigner) -> 'IResponseMode': ...

    async def deny(self) -> str: ...

    async def error(
        self,
        *,
        error: str,
        error_description: str | None = None,
        error_uri: str | None = None
    ) -> str:
        ...

    async def redirect(self, **kwargs: Any) -> fastapi.responses.RedirectResponse: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import contextlib

import pydantic

from oauthx.lib.protocols import IStorage


class BaseModel(pydantic.BaseModel):
    _storage: IStorage = pydantic.PrivateAttr()

    def attach(self, storage: IStorage):
        self._storage = storage
        return self

    @contextlib.asynccontextmanager
    async def atomic(self):
        try:
            yield
            await self.persist()
        except Exception:
            raise

    @contextlib.asynccontextmanager
    async def consume(self):
        try:
            yield
            await self.delete()
        except Exception:
            raise

    async def delete(self) -> None:
        await self._storage.delete(self)

    async def persist(self) -> None:
        await self._storage.persist(self)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import logging
from typing import Any
from typing import AsyncGenerator

import httpx

from oauthx.client.exceptions import AuthorizationRequired
from oauthx.client.protocols import IClient
from oauthx.client.protocols import IClientAuthorizationState
from oauthx.client.protocols import IClientStorage
from oauthx.client.protocols import IObtainedCredential


class AuthorizationCodeCredential(httpx.Auth):
    client: IClient | None
    credential: IObtainedCredential | None
    lock: asyncio.Lock = asyncio.Lock()
    logger: logging.Logger = logging.getLogger('uvicorn')
    name: str
    storage: IClientStorage[IClient, IClientAuthorizationState, IObtainedCredential]

    def __init__(self, storage: Any, name: str):
        self.storage = storage
        self.client = None
        self.credential = None
        self.name = name

    def get_credential_name(self) -> str:
        return self.name

    async def async_auth_flow(self, request: httpx.Request) -> AsyncGenerator[
        httpx.Request,
        httpx.Response,
    ]:
        if self.client is None:
            self.logger.info("Retrieving OAuth 2.x/OpenID Connect client (client: %s)", self.name)
            self.client = await self.storage.get_client(self.name)
        async with self.lock:
            if self.credential is None:
                self.logger.info("Retrieving credential (client: %s)", self.name)
                credential = await self.get_credential()
                if credential is None:
                    raise AuthorizationRequired
                self.credential = credential
        self.credential.add_to_request(request)
        response = yield request
        if response.status_code == 401:
            await self.refresh()
            self.credential.add_to_request(request)
            response = yield request
            if response.status_code == 401:
                await self.credential.destroy()
                self.credential = None
                self.logger.critical("Authorization required (client: %s)", self.name)
                raise AuthorizationRequired

    async def get_credential(self) -> IObtainedCredential | None:
        return  await self.storage.get_credential(self.get_credential_name())

    async def refresh(self) -> None:
        assert self.client is not None
        assert self.credential is not None
        async with self.lock:
            self.logger.info("Refreshing access token (client: %s)", self.name)
            await self.credential.refresh(self.client)
            await self.storage.persist(self.credential)


#!/usr/bin/env python3
# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from .models import ObtainedCredential
from .params import AuthorizationCodeGrant
from .params import Client
from .params import ClientStorage


class RedirectionRouter(fastapi.APIRouter):
    __module__: str = 'oauthx.client'

    def __init__(self, name: str):
        self.name = name
        super().__init__(
            dependencies=[fastapi.Depends(self.set_client)],
            include_in_schema=True
        )
        self.add_api_route(
            methods=['GET'],
            path='/authorize',
            name=f'oauth2.authorize.{self.name}',
            endpoint=self.authorize,
        )
        self.add_api_route(
            methods=['GET'],
            path='/callback',
            name=f'oauth2.callback.{self.name}',
            endpoint=self.callback,
        )

    async def authorize(
        self, request: fastapi.Request,
        client: Client,
        storage: ClientStorage
    ) -> fastapi.responses.RedirectResponse:
        state = await client.authorize(
            redirect_uri=str(request.url_for(f'oauth2.callback.{self.name}'))
        )
        await storage.persist(state)
        return state.redirect(fastapi.responses.RedirectResponse)

    async def callback(
        self,
        grant: AuthorizationCodeGrant,
        storage: ClientStorage
    ) -> None:
        if not grant.access_token:
            raise NotImplementedError
        credential = ObtainedCredential(
            client_id=self.name,
            name=self.name,
            access_token=grant.access_token,
            refresh_token=grant.refresh_token
        )
        await storage.persist(credential)

    def set_client(self, request: fastapi.Request) -> None:
        setattr(request.state, 'oauth_active_client', self.name)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .applicationclient import ApplicationClient
from .authorizationcodegrant import AuthorizationCodeGrant
from .bearertoken import BearerToken
from .client import Client
from .clientauthorizationstate import ClientAuthorizationState
from .clientstorage import ClientStorage
from .clientstorageimplementation import ClientStorageImplementation
from .provider import Provider
from .userinfo import UserInfo


__all__: list[str] = [
    'ApplicationClient',
    'AuthorizationCodeGrant',
    'BearerToken',
    'Client',
    'ClientAuthorizationState',
    'ClientStorage',
    'ClientStorageImplementation',
    'Provider',
    'UserInfo',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import datetime
import logging
from typing import cast
from typing import AsyncGenerator

from aiopki.ext import jose
from aiopki.ext.jose import JWT
import google.oauth2.id_token
import google.auth.transport.requests

import httpx


class DefaultServiceAccountAuth(httpx.Auth):
    audience: str | None = None
    bearer: str | None = None
    logger: logging.Logger = logging.getLogger('uvicorn')
    token: JWT | None = None

    def __init__(self, audience: str | None = None):
        self.audience = audience
        self.loop = asyncio.get_running_loop()

    async def async_auth_flow(
        self,
        request: httpx.Request
    ) -> AsyncGenerator[httpx.Request, httpx.Response]:
        now = datetime.datetime.now(datetime.timezone.utc)\
            - datetime.timedelta(seconds=30)
        if self.token is None or not self.token.validate_exp(now=now):
            self.bearer, self.token = await self.get_id_token(request)
        request.headers['Authorization'] = f'Bearer {self.bearer}'
        yield request

    async def get_id_token(self, request: httpx.Request) -> tuple[str, JWT]:
        audience = self.audience or f'{request.url.scheme}://{request.url.netloc.decode()}'
        self.logger.info(
            "Obtaining OIDC ID Token using default credentials"
            " (audience: %s)",
            audience
        )
        result = await self.loop.run_in_executor( # type: ignore
            None,
            google.oauth2.id_token.fetch_id_token, # type: ignore
            google.auth.transport.requests.Request(),
            audience
        )
        jws = jose.parse(result)
        return cast(str, result), jws.payload(JWT.model_validate)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import inspect
from typing import Any
from typing import TypeVar

from oauthx.lib import utils
from .config import Config
from .params import CurrentConfig


T = TypeVar('T')


class BaseTokenRevoker:
    __module__: str = 'oauthx.server.ref'
    config: Config

    @utils.class_property
    def __signature__(cls) -> inspect.Signature:
        return utils.merge_signatures([
            inspect.signature(cls.__init__),
            inspect.signature(cls.setup)
        ])

    def __init__(self, config: Config = CurrentConfig, **kwargs: Any) -> None:
        self.config = config
        self.setup(**kwargs)

    def setup(self, **kwargs: Any) -> None:
        pass

    async def revoke_hashed(self, token: str) -> None:
        raise NotImplementedError

    @functools.singledispatchmethod
    async def revoke(self, token: Any) -> None:
        raise NotImplementedError

    @revoke.register
    async def _(self, token: str) -> None:
        return await self.revoke_hashed(token)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from typing import Literal

import pydantic


class RefreshTokenPolicy(pydantic.BaseModel):
    current: str = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(48)
    )
    
    use: Literal['once', 'rolling', 'fixed'] = pydantic.Field(
        default=...
    )

    def rotate(self) -> None:
        if self.use in {'once'}:
            self.current = secrets.token_urlsafe(48)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
import pathlib
import re
import os
from typing import Any

import aiofiles
import jinja2
import yaml
from canonical.protocols import IRepository

from oauthx.client.models import Client


class LocalClientRepository(IRepository[Client]):
    base_dir: pathlib.Path = pathlib.Path('etc/upstream.d')
    logger: logging.Logger = logging.getLogger('uvicorn')
    pattern: re.Pattern[str] = re.compile('^[a-zA-Z0-9]([a-zA-Z0-9.]+)[a-zA-Z0-9]$')

    def construct_path(self, client_id: str) -> pathlib.Path | None:
        if not self.pattern.match(str(client_id) or ''):
            return None
        return self.base_dir.joinpath(f'{client_id}.yaml')

    async def delete(self, instance: Client | int | str) -> None:
        raise NotImplementedError

    async def exists(self, pk: str) -> bool:
        path = self.construct_path(pk)
        return path is not None and path.exists()

    async def get(self, pk: Any) -> Client | None:
        if not await self.exists(pk):
            return
        return await self.open(pk)

    async def one(self, pk: Any) -> Client:
        return await self.open(pk)

    async def open(self, pk: str) -> Client:
        path = self.construct_path(pk)
        if path is None or not path.exists():
            raise self.DoesNotExist
        async with aiofiles.open(path) as f: # type: ignore
            t = jinja2.Template(
                source=await f.read(), # type: ignore
                undefined=jinja2.StrictUndefined,
                variable_start_string='${',
                variable_end_string='}'
            )
            c = {'env': dict(os.environ)}
            params = yaml.safe_load(t.render(c)) # type: ignore
        obj = Client.model_validate(params)
        return obj.attach(self)

    async def persist(self, instance: Client) -> None:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .authorizationresponse import AuthorizationResponse
from .authorizationrequestparameters import AuthorizationRequestParameters
from .baseauthorizationrequest import BaseAuthorizationRequest
from .client import Client
from .clientauthorizationstate import ClientAuthorizationState
from .obtainedcredential import ObtainedCredential
from .provider import Provider
from .redirectionparameters import RedirectionParameters


__all__: list[str] = [
    'AuthorizationResponse',
    'AuthorizationRequestParameters',
    'BaseAuthorizationRequest',
    'Client',
    'ClientAuthorizationState',
    'ObtainedCredential',
    'Provider',
    'RedirectionParameters',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Protocol
from typing import TypeVar


T = TypeVar('T')


class IClientAuthorizationState(Protocol):
    __module__: str = 'oauthx.client.protocols'

    @property
    def pk(self) -> str:
        ...

    def annotation(self, name: str, decoder: Callable[[Any], T] = lambda x: x) -> T | None: ...
    def get_authorize_url(self) -> str: ...
    def redirect(self, request_factory: Callable[..., T]) -> T: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Iterable
from typing import Protocol
from typing import TypeVar

import httpx

from oauthx.lib.models import ObtainedGrant
from oauthx.lib.types import PrompType
from oauthx.lib.types import ResponseType
from oauthx.lib import utils
from .iclientauthorizationstate import IClientAuthorizationState


T = TypeVar('T')


class IClient(Protocol):
    __module__: str = 'oauthx.client.protocols'

    @property
    def pk(self) -> str:
        ...

    async def authorize(
        self,
        *,
        redirect_uri: str | None = None,
        scope: Iterable[str] | None = None,
        response_type: ResponseType | None = None,
        state_class: type[T] = IClientAuthorizationState,
        prompt: Iterable[PrompType] = [],
        annotations: dict[str, str] | None = None,
        **params: Any
    ) -> T:
        ...

    async def authorization_code(
        self,
        params: Any,
        state: Any,
        http: httpx.AsyncClient | None = None,
        resources: str | list[str] | None = None
    ) -> ObtainedGrant: ...

    @utils.http
    async def refresh(
        self,
        http: httpx.AsyncClient,
        refresh_token: str,
        scope: Iterable[str] | None = None,
        resources: list[str] | None = None,
        resource: str | None = None
    ) -> ObtainedGrant:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .application import create_application
from .auth import AuthorizationCodeCredential
from .clientresolver import ClientResolver
from .settings import ClientSettings
from .defaultclient import DefaultClient
from .localclientrepository import LocalClientRepository
from .models import *
from .redirectionrouter import RedirectionRouter
from . import params
from . import protocols


__all__: list[str] = [
    'create_application',
    'AuthorizationCodeCredential',
    'Client',
    'ClientSettings',
    'ClientResolver',
    'DefaultClient',
    'LocalClientRepository',
    'RedirectionRouter',
]


def setup_dependencies(
    storage_class: Any
) -> list[Any]:
    return [
        params.ClientStorageImplementation(storage_class)
    ]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
import pytest

from oauthx.client import Client as ConsumingClient
from .types import AuthorizeFactory


@pytest.fixture(scope='function')
def scope() -> list[str]:
    return ['offline_access']


@pytest.mark.asyncio
async def test_refresh_token_is_not_issued_without_offline_access(
    agent: httpx.AsyncClient,
    client: ConsumingClient,
    authorize_request: AuthorizeFactory
):
    _, response, state = await authorize_request()
    assert response
    assert state
    grant = await client.authorization_code(response, state, http=agent)
    assert grant.refresh_token is None


@pytest.mark.asyncio
async def test_refresh_token_is_issued_with_offline_access(
    agent: httpx.AsyncClient,
    client: ConsumingClient,
    authorize_request: AuthorizeFactory
):
    _, response, state = await authorize_request(
        scope={'offline_access'}
    )
    assert response
    assert state
    grant = await client.authorization_code(response, state, http=agent)
    assert grant.refresh_token is not None


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Iterable
from typing import Literal
from typing import TypeVar

import pydantic
from aiopki.ext.jose import OIDCToken
from canonical import DomainName
from canonical import EmailAddress
from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema


__all__: list[str] = ['ClientACL']


T = TypeVar('T')


class ClientACL:
    __module__: str = 'oauthx.server.models'
    conditions: list['Condition']

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        schema = core_schema.union_schema([
            core_schema.chain_schema([
                core_schema.list_schema(),
                core_schema.no_info_plain_validator_function(cls.fromiterable)
            ]),
            core_schema.chain_schema([
                core_schema.none_schema(),
                core_schema.no_info_plain_validator_function(cls.null)
            ])   
        ])
        return core_schema.json_or_python_schema(
            json_schema=schema,
            python_schema=core_schema.union_schema([
                schema,
                core_schema.is_instance_schema(cls)
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(lambda x: x.serialize())
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.union_schema([
            core_schema.list_schema(),
            core_schema.none_schema()
        ]))

    @classmethod
    def fromiterable(cls, v: Iterable[str]) -> 'ClientACL':
        conditions: list[Condition] = []
        for condition in v:
            kind, value = str.split(condition, ':', 1)
            conditions.append(Condition.model_validate({
                'kind': kind,
                'value': value
            }))

        return cls(conditions=conditions)

    @classmethod
    def null(cls, v: None = None) -> 'ClientACL':
        return cls(conditions=[Condition.model_validate({'kind': 'null'})])

    def __init__(
        self,
        conditions: Iterable['Condition'] | None = None
    ) -> None:
        self.conditions = list(conditions or [])

    def serialize(self) -> list[str] | None:
        return [
            str(condition) for condition in self.conditions
        ]

    def has_access(self, userinfo: OIDCToken) -> bool:
        return any([x.has_access(userinfo) for x in self.conditions])


class BaseCondition(pydantic.BaseModel):
    kind: Any
    value: Any

    def get_value(self) -> str:
        return self.value

    def has_access(self, userinfo: OIDCToken) -> bool:
        raise NotImplementedError

    def __str__(self) -> str:
        return f'{self.kind}:{self.get_value()}'


class DomainCondition(BaseCondition):
    kind: Literal['domain']
    value: DomainName

    def has_access(self, userinfo: OIDCToken) -> bool:
        return userinfo.email is not None and userinfo.email.domain == self.value


class EmailCondition(BaseCondition):
    kind: Literal['email']
    value: EmailAddress

    def has_access(self, userinfo: OIDCToken) -> bool:
        return userinfo.email is not None and userinfo.email == self.value


class NullCondition(BaseCondition):
    kind: Literal['null']
    value: str = 'allUsers'

    def has_access(self, userinfo: OIDCToken) -> bool:
        return True


class Condition(pydantic.RootModel[DomainCondition | EmailCondition | NullCondition]):

    def has_access(self, userinfo: OIDCToken) -> bool:
        return self.root.has_access(userinfo)

    def __hash__(self) -> int:
        return hash((self.root.get_value(), self.root.kind))

    def __str__(self) -> str:
        return str(self.root)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

import pydantic

from oauthx.server.params import ContentEncryptionKey
from .basedatastorestorage import BaseDatastoreStorage
from .params import DatastoreClient


T = TypeVar('T', bound=pydantic.BaseModel)


class DatastoreStorage(BaseDatastoreStorage):

    def __init__( # type: ignore
        self,
        *,
        client: DatastoreClient, # type: ignore
        key: ContentEncryptionKey
    ) -> None:
        self.client = client
        self.key = key


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Callable
from typing import Protocol

import httpx
from aiopki.ext.jose import JOSEObject
from fastapi.security import HTTPBasicCredentials


class IClient(Protocol):
    __module__: str = 'oauthx.lib.protocols'

    @property
    def id(self) -> str:
        ...

    def is_confidential(self) -> bool:
        """Return a boolean indicating if the client is confidential."""
        ...

    def update_key(self, update: Callable[[bytes], None]):
        """Update the clients' key that is used to secure its
        security credentials, such as authorization codes and
        refresh tokens. Updating the key effectively revokes
        all security credentials for the client.
        """
        ...

    async def authenticate(self, credential: HTTPBasicCredentials | JOSEObject | str | None) -> None:
        """Authenticate the client using the given credential.
        """
        ...

    async def userinfo(self, access_token: str, http: httpx.AsyncClient | None = None) -> None:
        """Query the providers' UserInfo endpoint to obtain information
        about the resource owner.
        """
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import httpx
import pydantic
from canonical import PersistedModel

from oauthx.client.exceptions import AuthorizationRequired
from oauthx.client.protocols import IClient
from oauthx.lib.exceptions import Error


class ObtainedCredential(PersistedModel):
    client_id: str = pydantic.Field(
        default=...
    )

    name: str = pydantic.Field(
        default=...
    )
    
    access_token: str = pydantic.Field(
        default=...
    )
    
    refresh_token: str | None = pydantic.Field(
        default=None
    )

    @property
    def pk(self) -> str:
        return self.name

    def add_to_request(self, request: httpx.Request) -> None:
        request.headers['Authorization'] = f'Bearer {self.access_token}'

    async def destroy(self) -> None:
        await self.delete()

    async def refresh(self, client: IClient) -> None:
        if self.refresh_token is None:
            raise AuthorizationRequired
        try:
            grant = await client.refresh(self.refresh_token)
        except Error:
            raise AuthorizationRequired
        if grant.refresh_token is None:
            raise AuthorizationRequired
        self.access_token = str(grant.access_token)
        self.refresh_token = grant.refresh_token


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .accesstoken import AccessToken
from .idtoken import IDToken
from .idtokenbearer import IDTokenBearer
from .requestaccesstoken import RequestAccessToken
from .requestsubject import RequestSubject
from .requiredscope import RequiredScope


__all__: list[str] = [
    'AccessToken',
    'IDToken',
    'IDTokenBearer',
    'RequestAccessToken',
    'RequestSubject',
    'RequiredScope',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import InvalidToken
from oauthx.resource import AccessToken
from oauthx.server.params import Storage
from oauthx.server.protocols import IRegisteredClient


async def get(at: AccessToken, storage: Storage) -> IRegisteredClient:
    client = await storage.get(at.client_id)
    if client is None:
        raise InvalidToken("The client is disabled.")
    return client


Client: TypeAlias = Annotated[IRegisteredClient, fastapi.Depends(get)]




# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.resource.models import Request
from oauthx.lib import RFC9068AccessToken
from oauthx.lib.exceptions import InvalidRequest


__all__: list[str] = ['AccessToken']


async def get(request: Request) -> RFC9068AccessToken | None:
    at = getattr(request, 'access_token', None)
    if at is None:
        raise InvalidRequest("The Authorization header is required.")
    return at


AccessToken: TypeAlias = Annotated[
    RFC9068AccessToken,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi


__all__: list[str] = ['ClientStorageImplementation']


def ClientStorageImplementation(cls: Any) -> Any:
    def f(request: fastapi.Request, impl: cls = fastapi.Depends(cls)):
        setattr(request.state, 'oauth_client_storage', impl)
        return impl
    return fastapi.Depends(f)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.params import Logger
from oauthx.server.models import AuthorizationRequest as AuthorizationRequestModel
from oauthx.server.params import ObjectFactory
from .client import Client
from .pushedauthorizationrequest import PushedAuthorizationRequest
from .redirecturi import RedirectURI
from .scope import Scope
from .targetresources import TargetResources


__all__: list[str] = [
    'AuthorizationRequest'
]


async def get(
    factory: ObjectFactory,
    logger: Logger,
    request: fastapi.Request,
    client: Client,
    redirect_uri: RedirectURI,
    resources: TargetResources,
    par: PushedAuthorizationRequest,
    scope: Scope,
) -> AuthorizationRequestModel:
    if par is None:
        par = await factory.request(
            client=client, # type: ignore
            request=request,
            redirect_uri=redirect_uri,
            resources=resources,
            scope=scope
        )
        await par.persist()
        logger.debug(
            "Persisted authorization request parameters (client: %s, urn: %s)",
            client.id, par.request_uri
        )
    return par

AuthorizationRequest: TypeAlias = Annotated[
    AuthorizationRequestModel,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import secrets
from typing import Any
from typing import Callable
from typing import Iterable
from typing import Literal
from typing import MutableMapping
from typing import TypeAlias

import pydantic

from oauthx.lib.protocols import IStorage
from oauthx.lib.types import TargetResource
from oauthx.server.protocols import IUserInfoContributor
from oauthx.server.types import AuthorizationKey
from .basemodel import BaseModel
from .refreshtokenpolicy import RefreshTokenPolicy
from .subjectkey import SubjectKey


AuthorizedTokenType: TypeAlias = Literal[
    'urn:ietf:params:oauth:token-type:access_token',
    'urn:ietf:params:oauth:token-type:refresh_token',
    'urn:ietf:params:oauth:token-type:id_token',
]


class Authorization(BaseModel):
    """A :class:`Authorization` object represents the authorization granted
    by the resource owner through the authorization endpoint.
    """
    authorized: datetime.datetime = pydantic.Field(
        default_factory=lambda: datetime.datetime.now(datetime.timezone.utc)
    )

    code: str | None = pydantic.Field(
        default=None
    )

    client_id: str = pydantic.Field(
        default=...
    )

    id: int = pydantic.Field(
        default=...
    )
    
    issuer: str = pydantic.Field(
        default=...
    )

    lifecycle: Literal['GRANTED', 'ISSUED'] = pydantic.Field(
        default='ISSUED'
    )

    prompted: bool = pydantic.Field(
        default=False
    )

    refresh_token: RefreshTokenPolicy | None = pydantic.Field(
        default=None
    )
    
    resources: set[TargetResource] = pydantic.Field(
        default_factory=set
    )

    scope: set[str] = pydantic.Field(
        default_factory=set
    )

    sub: SubjectKey = pydantic.Field(
        default=...
    )
    
    token_types: set[AuthorizedTokenType] = pydantic.Field(
        default=...
    )
    
    userinfo: dict[str, Any] = pydantic.Field(
        default={}
    )

    @property
    def pk(self) -> AuthorizationKey:
        return AuthorizationKey(self.id)

    def model_post_init(self, _: Any) -> None:
        if 'offline_access' in self.scope:
            self.token_types.add('urn:ietf:params:oauth:token-type:refresh_token')

    def allows_scope(self, scope: Iterable[str]) -> bool:
        return set(scope) <= self.scope

    def allows_target(self, target: TargetResource | Iterable[TargetResource]) -> bool:
        if isinstance(target, TargetResource):
            target = {target}
        return set(target) <= self.resources

    def is_consumed(self) -> bool:
        return self.lifecycle == 'ISSUED'

    def contribute(self, contributor: IUserInfoContributor) -> None:
        contributor.contribute_to_userinfo(self.userinfo)

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        userinfo.update(self.userinfo)

    def grants_id_token(self) -> bool:
        return 'urn:ietf:params:oauth:token-type:id_token' in self.token_types

    def grants_refresh_token(self) -> bool:
        return 'urn:ietf:params:oauth:token-type:refresh_token' in self.token_types

    def update_key(self, update: Callable[[bytes], None]) -> None:
        return

    async def consume(self, storage: IStorage) -> None:
        self.lifecycle = 'ISSUED'
        await storage.persist(self)

    async def issue_authorization_code(self) -> str:
        if self.code is not None:
            raise TypeError(
                "Authorization code is already issued for this grant."
            )
        self.code = secrets.token_urlsafe(32)
        return self.code

    async def issue_refresh_token(self) -> str:
        if not self.grants_refresh_token():
            raise TypeError(
                "Can not issue a refresh token if the `offline_access` scope is "
                "not granted by the resource owner."
            )
        if not self.refresh_token:
            self.refresh_token = RefreshTokenPolicy(
                use='once'
            )
        else:
            self.refresh_token.rotate()
        await self.persist()
        return self.refresh_token.current


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from .types import AuthorizeFactory


@pytest.mark.asyncio
@pytest.mark.parametrize("resource", [
    'foo/bar/baz',
    ['foo/bar/baz'],
    ['foo/bar/baz', 'baz/bar/foo']
])
async def test_resource_must_be_absolute_uri(
    resource: str | list[str],
    authorize_request: AuthorizeFactory
):
    _, response, _ = await authorize_request(
        resource=resource,
        expect_error='invalid_target',
        expect_status=302
    )
    assert response.is_error()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import MutableMapping
from typing import Protocol

import httpx
from aiopki.ext.jose import JOSEObject
from aiopki.ext.jose import OIDCToken
from fastapi.security import HTTPBasicCredentials

from oauthx.lib.types import GrantType
from oauthx.lib.types import RedirectURI
from oauthx.lib.types import ResponseType
from .iclient import IClient


class IRegisteredClient(Protocol):
    __module__: str = 'oauthx.lib.protocols'
    Key: ClassVar[type[IClient.Key]] = IClient.Key

    @property
    def id(self) -> str:
        ...

    def allows_delegation_to(self, userinfo: OIDCToken) -> bool:
        """Return a boolean indicating if the client allows delegation to
        the given identity.
        """
        ...

    def allows_response_type(self, response_type: ResponseType | None) -> bool:
        """Return a boolean indicating if the client allows the given
        response type.
        """
        ...

    def allows_scope(self, scope: set[str]) -> bool:
        """Return a boolean indicating if the client allows the
        given scope.
        """
        ...

    def can_redirect(self, redirect_uri: RedirectURI | None) -> bool:
        """Return a boolean indicating if the client allows
        redirection to the given `redirect_uri`.
        """
        ...

    def can_grant(self, grant_type: GrantType | None) -> bool:
        """Return a boolean indicating if the client allows the given
        `grant_type`.
        """
        ...

    def contribute_to_event(self, data: MutableMapping[str, Any]) -> None:
        """Adds properties to an event that involves the client."""
        ...

    def contribute_to_userinfo(self, userinfo: MutableMapping[str, Any]) -> None:
        """Adds properties to a :class:`UserInfo` instance."""
        ...

    def get_default_redirect_uri(self) -> RedirectURI:
        """Return a string containing the default redirect URI for the client,
        if any.
        """
        ...

    def get_display_name(self) -> str:
        """Return a string containing the display name of the client as
        shown in the user interface.
        """
        ...

    def get_logo_url(self) -> str:
        """Return a string containing the logo URL of the client."""
        ...

    def get_sector_identifier(self) -> str:
        """Return a string containing the :term:`Sector Identifier` of the
        client.
        """
        ...

    def has_redirection_endpoints(self) -> bool:
        """Return a boolean indicating if the client has configured any
        redirection endpoints.
        """
        ...

    def is_confidential(self) -> bool:
        """Return a boolean indicating if the client is confidential."""
        ...

    def must_push(self) -> bool:
        """Return a boolean indicating if the client is required to
        push authorization requests.
        """
        ...

    def requires_state(self) -> bool:
        """Return a boolean indicating if the client enforces the use
        of the ``state`` parameter in authorization requests.
        """
        ...

    def resource_owner(self, subject: Any) -> Any:
        """Create a resource owner identifier."""
        ...

    def update_key(self, update: Callable[[bytes], None]):
        """Update the clients' key that is used to secure its
        security credentials, such as authorization codes and
        refresh tokens. Updating the key effectively revokes
        all security credentials for the client.
        """
        ...

    async def authenticate(self, credential: HTTPBasicCredentials | JOSEObject | str | None) -> None:
        """Authenticate the client using the given credential.
        """
        ...

    async def userinfo(self, access_token: str, http: httpx.AsyncClient | None = None) -> None:
        """Query the providers' UserInfo endpoint to obtain information
        about the resource owner.
        """
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import Any
from typing import TypeAlias

import fastapi
from aiopki.ext.jose import OIDCToken

from .contentencryptionkey import ContentEncryptionKey
from .currentsubject import CurrentSubject
from .objectfactory import ObjectFactory


__all__: list[str] = [
    'RequestUserInfo',
    'RequestUserInfoDependency',
]


async def get(
    factory: ObjectFactory,
    subject: CurrentSubject,
    cek: ContentEncryptionKey,
) -> OIDCToken | None:
    if subject is None or not subject.is_authenticated():
        return None
    await subject.decrypt_keys(cek)
    return await factory.userinfo(
        subject=subject,
        scope={'profile', 'email'},
        aud='self'
    )


RequestUserInfoDependency: Any = fastapi.Depends(get)

RequestUserInfo: TypeAlias  = Annotated[None, RequestUserInfoDependency]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
from aiopki.ext.jose import OIDCToken

from oauthx.lib.params import Logger
from oauthx.server.params import ContentEncryptionKey
from oauthx.server.params import CurrentSubject
from oauthx.server.params import ObjectFactory
from oauthx.server.params import RequestSession
from oauthx.server.types import LoginRequired
from oauthx.server.types import UnauthorizedAccount
from .authorizationrequest import AuthorizationRequest
from .client import Client
from .responsemode import ResponseMode


__all__: list[str] = [
    'UserInfo'
]


async def get(
    factory: ObjectFactory,
    logger: Logger,
    key: ContentEncryptionKey,
    request: fastapi.Request,
    session: RequestSession,
    client: Client,
    subject: CurrentSubject,
    response_mode: ResponseMode,
    par: AuthorizationRequest
) -> OIDCToken:
    if subject is None:
        logger.debug(
            "Request is not authenticated (client: %s, request: %s)",
            client.id, par.request_uri
        )
        raise LoginRequired(
            client=client,
            next_url=par.get_authorize_url(request),
            deny_url=await response_mode.deny(),
            a=par.id
        )
    await subject.decrypt_keys(key)
    userinfo = await factory.userinfo(
        subject=subject,
        contributors=[session, client]
    )
    if not client.allows_delegation_to(userinfo):
        raise UnauthorizedAccount({
            'client_name': client.get_display_name(),
            'client_logo': client.get_logo_url(),
            'logout_url': request.url_for('user.logout')\
                .include_query_params(n=par.get_authorize_url(request)),
            'deny_url': await response_mode.deny()
        })
    return userinfo

UserInfo: TypeAlias = Annotated[
    OIDCToken,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Annotated
from typing import TypeAlias

import fastapi


__all__: list[str] = ['NextURL']


def is_allowed(request: fastapi.Request, url: str | None) -> bool:
    try:
        p = urllib.parse.urlparse(url)
        return all([
            p.scheme == request.url.scheme,
            p.netloc == request.url.netloc
        ])
    except Exception:
        return False

def get(
    request: fastapi.Request,
    next_url: str | None = fastapi.Query(
        default=None,
        alias='n',
        title="Next URL"
    ),
) -> str | None:
    if not is_allowed(request, next_url):
        raise NotImplementedError
    return next_url


NextURL: TypeAlias = Annotated[str | None, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

from aiopki.ext.jose import OIDCToken

from oauthx.resource.params import IDTokenBearer


__all__: list[str] = ['GoogleServiceAccount']


GoogleServiceAccount: TypeAlias = Annotated[
    OIDCToken | None,
    IDTokenBearer('https://accounts.google.com')
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.client.protocols import IClient
from .clientstorage import ClientStorage


__all__: list[str] = ['Client']


async def get(
    request: fastapi.Request,
    storage: ClientStorage
) -> IClient:
    return await storage.get_client(getattr(request.state, 'oauth_active_client'))


Client: TypeAlias  = Annotated[IClient, fastapi.Depends(get)]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Annotated
from typing import TypeAlias

import fastapi

from oauthx.lib.exceptions import UnknownClient
from oauthx.server.params import Storage
from oauthx.server.models import ClientKey
from oauthx.server.protocols import IRegisteredClient
from oauthx.server.request import Request
from oauthx.server.types import MissingRedirectURI
from .query import CLIENT_ID


__all__: list[str] = ['Client']


async def get(
    request: Request,
    storage: Storage,
    client_id: str = CLIENT_ID,
) -> IRegisteredClient:
    client = cast(IRegisteredClient | None, await storage.get(ClientKey(client_id)))
    if client is None:
        raise UnknownClient("The client_id parameter specifies an invalid client.")
    if not client.has_redirection_endpoints():
        raise MissingRedirectURI
    request.set_client(client.root) # type: ignore
    return client.root # type: ignore


Client: TypeAlias = Annotated[
    IRegisteredClient,
    fastapi.Depends(get)
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Annotated
from typing import TypeAlias

import fastapi
import pydantic

from oauthx.lib.exceptions import InvalidRequest
from oauthx.lib.params import Logger
from oauthx.server.request import Request
from oauthx.server.params import IssuerIdentifier
from oauthx.server.params import ObjectFactory
from oauthx.server.params import TokenSigner
from oauthx.server.types import InvalidAuthorizationRequest
from oauthx.server.types import InvalidResponseType
from oauthx.server.types import IResponseMode
from oauthx.server.types import UnauthorizedClient
from .client import Client
from .query import RESPONSE_MODE
from .redirecturi import RedirectURI
from .responsetype import ResponseType
from .state import State


__all__: list[str] = [
    'ResponseMode'
]


async def get(
    factory: ObjectFactory,
    signer: TokenSigner,
    logger: Logger,
    iss: IssuerIdentifier,
    request: Request,
    client: Client,
    response_type: ResponseType,
    redirect_uri: RedirectURI,
    state: State,
    response_mode: str | None = RESPONSE_MODE,
) -> IResponseMode:
    try:
        obj = await factory.response_mode(
            iss=iss,
            client=client, # type: ignore
            response_type=response_type,
            response_mode=response_mode,
            redirect_uri=redirect_uri,
            state=state
        )
        request.response_mode = obj.with_signer(signer) # type: ignore
    except pydantic.ValidationError:
        raise InvalidAuthorizationRequest(
            error='invalid_request',
            allow_redirect=True,
            redirect_uri=redirect_uri,
            context={'client_name': client.get_display_name()}
        )
    if not client.allows_response_type(obj.response_type):
        raise InvalidResponseType
    if not client.can_grant(obj.grants()):
        logger.debug(
            "Client does not allow the requested grant (client: %s)",
            client.id
        )
        raise UnauthorizedClient
    if client.requires_state() and not state:
        raise InvalidRequest("The client requires the use of the state parameter.")
    return obj.with_signer(signer)

ResponseMode: TypeAlias = Annotated[
    IResponseMode,
    fastapi.Depends(get)
]